/*
 *	esc-if.c --- "if" filter for ESC/P, invoked by lpd.
 *
 *			Copyright (C) 1994,1996 Koh-ichi Ito
 */

#include <stdio.h>
#include <syslog.h>
#include <signal.h>

#define WIDTH	 80	/* usually overridden by printcap */
#define LENGTH	 66	/* usually overridden by printcap */
#ifndef L_MARGIN
#define L_MARGIN  0	/* left margine */
#endif
#define INDENT	  0
#define TABS	  8

#define BS	0x08
#define TAB	0x09
#define FF	0x0c
#define ESC	0x1b
#define FS	0x1c

#define SI	printf("%c.", FS)
#define SO	printf("%c&", FS)

#define SJIS	1
#define JIS	2
#define EUC	3

void	preamble();
void	postamble();
void	deadcopy();
void	abort();
void	sjis2jis();

struct args {
	int	c;	/* include ctrl chrs */
	int	width;
	int	length;
	int	indent;
	char	*name;
	char	*host;
	char	*acct;
};

char	*MyName;
struct args	Args = {
		0, WIDTH, LENGTH, INDENT, NULL, NULL, NULL};
int	Page;
static char	Header[] = "$Header: /home/lucy/kohi/src/esc-if/RCS/esc-if.c,v 1.5 1997/10/27 13:55:14 kohi Exp $";

main(argc, argv)

	int	argc;
	char	**argv;
{
	signal(SIGINT, abort);
	MyName = *argv;
	while (--argc) {
		if (**++argv != '-') {
			break;
		}
		switch (*++*argv) {
		case 'c':
			Args.c = 1;
			break;
		case 'w':
			Args.width = atoi(++*argv);
			break;
		case 'l':
			Args.length = atoi(++*argv);
			break;
		case 'i':
			Args.indent = atoi(++*argv);
			break;
		case 'n':
			Args.name = *++argv;
			argc--;
			break;
		case 'h':
			Args.host = *++argv;
			argc--;
			break;
		default:
			break;
		}
	}
	if (argc) {
		Args.acct = *argv;
	}
	if (Args.c) {
		deadcopy();
		record_acct();
		exit(0);
	}
	Page = 0;
	preamble();
	while (++Page, print_a_page() )
		;
	postamble();
	record_acct();
	exit(0);
}

int print_a_page()
{
	int	x, y, l, tabs;
	static int kanji = 0;
	int	c, c2;

	l = Args.width - Args.indent;
	for (y = 0; y < Args.length; y++) {
		for (x = 0; x < Args.indent + L_MARGIN; x++) {
			putchar(' ');
		}
		if (kanji) {
			SO;
		}
		x = 0;
		while (x < l) {
			c = getchar();
			switch (c) {
			case EOF:
				return 0;
			case FF:
				c = getchar();
				if (c == EOF) {
					return 0;
				}
				ungetc(c, stdin);
				if (kanji) {
					kanji = 0;
					SI;
				}
				putchar(FF);
				return 1;
			case '\031':
				/*
				 * when invoked as 'of',
				 * lpd request to stop 'of'
				 * by the sequence of '\031\1'.
				 */
				c = getchar();
				if (c == '\1') {
					putchar(FF);
					fflush(stdout);
					kill(getpid(), SIGSTOP);
				} else {
					ungetc(c, stdin);
					ungetc('\031', stdin);
				}
				break;
			case '\r':
				c = getchar();
				if (c == '\n') {
					goto EOLN;
				} else {
					ungetc(c, stdin);
					if (kanji) {
						kanji = 0;
						SI;
					}
					putchar('\r');
					for (x = 0; x < Args.indent
							+ L_MARGIN; x++) {
						putchar(' ');
					}
					continue;
				}
			case '\n':
				goto EOLN;
			case '\t':
				tabs = (int)( (x + TABS) / TABS) * TABS;
				if (l < tabs) {
					goto EOLN;
				}
				if (kanji) {
					kanji = 0;
					SI;
				}
				while (x < tabs) {
					putchar(' ');
					x++;
				}
				break;
			case ESC:
				c = getchar();
				if (c == '$') {
					kanji = JIS;
					getchar();	/* skip 1 char */
					SO;
					break;
				} else if (c == '(') {	/* ) */
					kanji = 0;
					getchar();	/* skip 1 char */
					SI;
					break;
				} else {
					putchar(ESC);
					putchar(c);
					x += 2;
					break;
				}
			default:
#ifdef USE_EUC
				if ( (kanji != EUC) && (c & 0x80) ) {
					kanji = EUC;
					SO;
				} else if ( (kanji == EUC)
					&& !(c & 0x80) ) {
					kanji = 0;
					SI;
				}
#else	/* USE_EUC */
				if ( (kanji != SJIS) && (c & 0x80) ) {
					kanji = SJIS;
					SO;
				} else if ( (kanji == SJIS)
					&& !(c & 0x80) ) {
					kanji = 0;
					SI;
				}
#endif	/* USE_EUC */
				if (kanji) {
					if (x == l - 1) {
						ungetc(c, stdin);
						goto EOLN;
					}
					c2 = getchar();
					if (kanji == SJIS) {
						sjis2jis(&c, &c2);
					}
					putchar(c & 0x7f);
					putchar(c2 & 0x7f);
					x += 2;
				} else {
					putchar(c);
					if (c == BS) {
						x--;
					} else {
						x++;
					}
				}
				break;
			}
		}
EOLN:
		if (kanji) {
			SI;
		}
		if (l <= x) {
			c = getchar();
			if (c != '\n') {
				ungetc(c, stdin);
			}
		}
		putchar('\n');
	}
	putchar(FF);
	return 1;
}

void preamble()
{
	printf("%c@", ESC);	/* reset printer */
	printf("%cM", ESC);	/* 12CPI */
	printf("%cS%c%c", FS, 3, 3);	/* spacing for 2Bytes char */
	SI;
}

void postamble()
{
	SI;
	putchar(FF);
	printf("%c@", ESC);	/* reset printer */
}

void deadcopy()
{
	int	c;

	while ( (c = getchar() ) != EOF) {
		putchar(c);
	}
}

void abort()
{
	record_acct();
	exit(0);
}

int record_acct()
{
	FILE	*fp;

	if (!Args.acct) {
		return 1;
	}
	if ( (fp = fopen(Args.acct, "a") ) == NULL) {
		fprintf(stderr, "%s: ", MyName);
		perror(Args.acct);
		return 0;
	}
#ifdef ACCT_BC
	fprintf(fp, "%s @ %s prints %d page(s)\n", Args.name,
		Args.host, Page);
#else
	fprintf(fp, "%d\t%s:%s\n", Page, Args.host, Args.name);
#endif	/* ACCT_BC */
	fclose(fp);
	return 1;
}

void sjis2jis(c1, c2)

	int	*c1, *c2;
{
	if (0x9f < *c1) {
		*c1 -= 0xb1;
	} else {
		*c1 -= 0x71;
	}
	*c1 = *c1 * 2 + 1;
	if (0x9f < *c2) {
		*c2 = *c2 - 0x7e;
		(*c1)++;
	} else if (0x7e < *c2) {
		*c2 -= 0x20;
	} else {
		*c2 -= 0x1f;
	}
}
